package jabba;

import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContext;
import org.omg.CosNaming.NamingContextHelper;

import org.omg.CosNaming.NamingContextPackage.NotFound;
import org.omg.CosNaming.NamingContextPackage.CannotProceed;
import org.omg.CosNaming.NamingContextPackage.InvalidName;

import org.omg.CORBA.ORB;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.net.URL;
import java.net.MalformedURLException;

import ABng_Core.View;
import ABng_Core.Sighting_Description;

import ABng_Server_Manager.TimeServer;

import Domain_Repository.User_Rep;
import Domain_Repository.Equipment_Rep;
import Domain_Repository.Location_Description_Rep;

import pl.edu.agh.ics.abngcore.Bootstrap_n;
import pl.edu.agh.ics.abngcore.ORB_wrapper;

import szTools.console.Console;
//import szTools.console.GraphicalConsole;
import szTools.console.TextConsole;
import szTools.console.NullConsole;

/**
 * The main class of the JABBA 2.0 application.
 * <BR><BR>Responsible for:<OL>
 * <LI> Initializing the ORB_wrapper.
 * <LI> Initializing the Bootstrap_n.
 * <LI> Binding to an ABng_Core.View instance.
 * <LI> Creating a Receiver object.
 * <LI> Creating a Viewer object.
 * <LI> Creating a Console object.
 * </OL>
 * @see Receiver
 * @see Viewer
 * @see Console
 */
public class Jabba {
  /** Shows messages from running objects. */
  Console console=null;    

  /** Receives messages from an ABng_Core.View object */ 
  Receiver receiver=null;  

  /** The object the application is currently binded to. */
  View view=null;        

  /** The User Repository */
  User_Rep user_repository;

  /** The Equipment Repository */
  Equipment_Rep equipment_repository;

  /** The Location Description Repository */
  Location_Description_Rep location_description_repository;

  /** Helps performing some operations on ABng_Core objects */
  Helper helper;           

  /** Current interface */
  JabbaInterface viewer;  

  /** Clock provides time synchronization. */
  Clock clock;

  /** Allows binding to different ORBs. */
  ORB_wrapper orb_wrapper;

  /** 
   * After this time from the last sighting message 
   * the user is considered to be absent. 
   */
  static final int ABSENT_SECONDS = 900;		// 15 min

////////////////////////// NASZE 

  static final int PERHAPS_SECONDS = 600;		// 10 min
  static final int MIGHT_SECONDS = 300;			// 5 min

///////////////////////// NASZE 

  static final int EQUIPMENT_NOT_PRESENT_SECONDS = 1800;	// 30 min

  /** Minimal time change */
  final static int TIME_GRID = 40;

  /** Time to check for absences after. */
  final static int CHECK_FOR_ABSENCES_MILLIS = 60000; // minuta

  final static Integer ABSENT_QUALITY = new Integer(-100);
  
  private final static String STD_CONSOLE="null";
  //  private final static String STD_NS_URL="http://iris.ics.agh.edu.pl/abs_data/NS.ior";

  /**
   * Constructs new Jabba object and binds it to the specified View.
   * @param  args      arguments to initialize ORB_wrapper with
   * @param  viewname  name of the View to be observed
   */
  private Jabba( String[] args, String viewname ) {
    // obiekt pomocniczy do podstawowych operacji na innych obiektach
    helper = new Helper();

    // tworzymy przyrzad do sledzenia 
    String consoleType = System.getProperty( "CONSOLE",STD_CONSOLE );
    if( consoleType.equals("graphical") ) {
//      console = new GraphicalConsole();
    }
    else if( consoleType.equals("text") ) {
      console = new TextConsole();
    }
    else {
      console = new NullConsole();
    }
    console.addWriter( this, "Jabba" );
    System.err.println( "Console created." );

    // inicjujemy orb oraz bootstrap
    orb_wrapper = new ORB_wrapper();
    orb_wrapper.init_application( args );

    // inicjujemy Bootstrap
    String NSurl = System.getProperty( "NS_URL",null );

    System.err.println( "URL: "+NSurl );

    if( NSurl!=null && NSurl.length() > 0 ) {
      URL url=null;
      try {
	url = new URL(NSurl);
      }
      catch( MalformedURLException muex ) {
	console.exception( this,muex );
	System.exit(1);
      }

      if( url==null ) 
	Bootstrap_n.init();
      else 
	Bootstrap_n.init(url);
    }
    else Bootstrap_n.init();

    System.err.println( "Bootstrap started." );

    // Inicjalizacja zegara
    TimeServer ts = Bootstrap_n.get_register_object();
    if( ts==null ) {
      System.err.println( "\n\n\nWarning! Cannot initialize clock.\n\n\n" );
    }
    clock = new Clock( ts );

    // pobieramy referencje obserwowanego widoku, user repository i equipment repository
    try{
      view = ABng_Core.ViewHelper.narrow( Bootstrap_n.resolve( 
        Bootstrap_n.concat_names( Bootstrap_n.CTX_ABng_Core_Public_Views(),
	  			  viewname ) ) );
      user_repository = Domain_Repository.User_RepHelper.narrow(
	Bootstrap_n.resolve( Bootstrap_n.SRV_User_Repository() ) );
      equipment_repository = Domain_Repository.Equipment_RepHelper.narrow(
        Bootstrap_n.resolve( Bootstrap_n.SRV_Equipment_Repository() ) );
      location_description_repository = Domain_Repository.Location_Description_RepHelper.narrow(
										       Bootstrap_n.resolve( Bootstrap_n.SRV_Location_Description_Repository() ) );
    }
    catch( NotFound ex ) {
      System.err.println( "(ex1): "+ex );
      System.exit(1);
    }
    catch( CannotProceed ex ) {
      System.err.println( "(ex2): "+ex );
      System.exit(1);
    }
    catch( InvalidName ex ) {
      System.err.println( "(ex3): "+ex );
      System.exit(1);
    }

    console.messageFrom( this,"Creating receiver..." );
    // tworzymy "odbieracz"
    if( view != null ) {
      String pom = System.getProperty( "EVENT_CHANNEL", "no" );
      boolean pomb = pom.equals("yes") || pom.equals("YES") || pom.equals( "Yes" ); 
      receiver = new Receiver( this, view, user_repository, equipment_repository, location_description_repository, pomb );
      System.err.println( "Receiver created." );
    }
    else System.err.println( "No view!" );

    console.messageFrom( this, "READY" );
  }

  /**
   * Registers new interface. Automatically deregisters the old one.
   * @param viewer new interface
   */
  void registerInterface( JabbaInterface viewer ) {
    // <TU DODAC> odlaczenie poprzedniego interfejsu </TU>
    this.viewer = viewer;
  }

  /**
   * Shows console if any exists. Does not create a new one.
   */
  void showConsole() {
    if( console != null ) { 
      console.display();
    }
  }

  /**
   * If there is any interface registered, shows it. Otherwise does nothing.
   */
  void showInterface() {
    if( viewer != null ) {
      viewer.display();
    }
    else
      System.err.println( "NULL Viewer!!!" );
  }

  /**
   * Exits the program.
   * @param code code to be returned to the system
   */
  void exit(int code) {
    // <TU DODAC> zatrzymanie watkow </TU>
    // <TU DODAC> powiadomienie konsoli </TU>
    System.exit(code);
  }

  /**
   * Helps performing some frequently needed operations on ABng_Core classes.
   * @see Jabba#helper
   */
  private class Helper {
    /** 
     * Creates a text description of ABng_Core.Sighting_Description
     * @param sd object to be described
     * @return human-understandable description
     */
    public String sighting_to_string( Sighting_Description sd ) {
      return "sighting( locatable_id: "+sd.my_locatable+" location: "+sd.my_location+ " )";
    }
  }

  /** 
   * Starts the JABBA 2.0 application.
   */
  public static void main( String args[] ) {
    Viewer viewer;
    Jabba jabba;  
    String vName;
    System.err.println( "Java started." );

    vName = System.getProperty( "VIEW_NAME","viewroom" );
    jabba = new Jabba( args, vName );

    // viewer sie sam zarejestruje w silniku
    Viewer.createFrame( jabba );

    // pokazanie okna glownego
    jabba.showInterface();

    // I juz! Gotowe!
    jabba.receiver.impl_is_ready();
  }

}




