// $Id: observed_i.cc,v 1.2 1998/07/29 13:14:53 mencnar Exp $
//
// $Source: /home/iris/grant/abng/CVS/sources/common/observed_i.cc,v $
//
// $Log: observed_i.cc,v $
// Revision 1.2  1998/07/29 13:14:53  mencnar
// time synchronisation,
// persistent observers,
// notify calls moved to another thread ( fixed deadlock)
// internal server names changed
//
// Revision 1.1  1998/06/16 12:55:00  mencnar
// Initial revision
//
//
// File:          observed.cc
// Author:        Jakub Szymaszek (jasz@ics.agh.edu.pl)
// Date Created:  02.02.98
//
// General Description:
//  Implementation of the template Observed_I class which implements
//  operations common for those observed ABng Repositories that notify
//  their observers when a new item is added to the repository.

#include "observed_i.h"

template <class Observer_Types>
Observed_I<Observer_Types>::Observed_I(
#ifdef FilePersistency
     const char *repository_name
#endif
) :
  Basic_Observed_I<Observer_Types>(
#ifdef FilePersistency
     repository_name
#endif
  ),
  my_add_observer_list(hash_observer) {}

template <class Observer_Types>
Observed_I<Observer_Types>::~Observed_I() {

  // Abort all add observers.
  // Construct iterator.
 Observer_List_Iterator iterator(my_add_observer_list);
 
  // Notify all add observers that the observed quits.
  while (iterator()) {
    try {
      iterator.key()->abort();
    }
    catch (...) {}
  } 
}

template <class Observer_Types>
Basic_Observed_I<Observer_Types>::Observers *
  Observed_I<Observer_Types>::add_observer_list (
  ENV_DEF1(env)) 
  throw (CORBA::SystemException) {

  auto_lock local_lock(*this);

  int observers_number = my_add_observer_list.entries();
 
  Observer_List_Iterator iterator(my_add_observer_list);
 
  // Initialize the sequence for collected observers.
  Observers* collected_observers =
    new Observers(observers_number);
  collected_observers->length(observers_number);
  for (int i = 0; iterator(); i++){
    // Current observer reference 
    const Observer_ptr obs_p = iterator.key();
    // Copy the observer reference.
    (*collected_observers)[i] = Observer::_duplicate(obs_p);
  }
  return collected_observers;
}

template <class Observer_Types>
void Observed_I<Observer_Types>::attach_add_observer (
  Observer_ptr obs
  ENV_DEF(env))
  throw (CORBA::SystemException) {

  auto_lock local_lock(*this);
  private_attach_add_observer(obs USE_ENV(env));
}

template <class Observer_Types>
void Observed_I<Observer_Types>::detach_add_observer (
  Observer_ptr obs
  ENV_DEF(env)) 
     throw (CORBA::SystemException, 
	    Domain_Repository::Unknown_Observer) {

  auto_lock local_lock( *this);
  private_detach_add_observer(obs USE_ENV(env));
}

template <class Observer_Types>
void Observed_I<Observer_Types>::private_attach_add_observer (
  Observer_ptr obs
  ENV_DEF(env))
  throw (CORBA::SystemException) {

  if (CORBA::is_nil( obs)) {
    return;
  }

#ifdef DEBUG
  cerr << "Observed_I::attach_add_observer " << (int) obs << endl;
#endif
  add_proxy(obs);
  if (!my_add_observer_list.contains(obs)) {
#ifdef DEBUG
    cerr << "New add observer" << endl;
#endif
    my_add_observer_list.insert(obs);
  }
}

template <class Observer_Types>
void Observed_I<Observer_Types>::private_detach_add_observer (
  Observer_ptr obs
  ENV_DEF(env)) {

  if ( CORBA::is_nil(obs)) { 
    throw Domain_Repository::Unknown_Observer();
  }
  add_proxy(obs);

  // Check if observer is in the add observer list
  if (!my_add_observer_list.contains(obs))
    // Unknown observer
    throw Domain_Repository::Unknown_Observer();
  else {
    // remove this observer from the list
    my_add_observer_list.remove(obs);
    //    CORBA::release(obs);
  }
  Basic_Observed_I<Observer_Types>::remove_proxy(obs);
}

template <class Observer_Types>
void Observed_I<Observer_Types>::remove_proxy(
  Observed_I<Observer_Types>::Observer_ptr obs) {

  if(my_add_observer_list.contains(obs)) return;
  Basic_Observed_I<Observer_Types>::remove_proxy(obs);

}

template <class Observer_Types>
void Observed_I<Observer_Types>::save_observers_to_stream( ostream &ostr) {

  Basic_Observed_I<Observer_Types>::save_observers_to_stream( ostr);
  ostr << endl;
  ostr << my_add_observer_list.entries() << endl;
  
  Observer_List_Iterator iterator( my_add_observer_list);
  while (iterator()){
    CORBA::String_var tmpstr=
      Bootstrap_base::orb->object_to_string( iterator.key());
    ostr << ((const char *) tmpstr) << endl;
  }
}

template <class Observer_Types>
void Observed_I<Observer_Types>::load_observers_from_stream( istream &ifs) {
  
  Basic_Observed_I<Observer_Types>::load_observers_from_stream( ifs);
  if (!ifs) return;
  int nof_obs;
  ifs >> nof_obs;
  if (!ifs) return;
  for (int i=0; i<nof_obs; i++ ) {
    RWCString tmpstr;
    ifs >> ws >> tmpstr;
    try {
      CORBA::Object_var tmpobj=
	Bootstrap_base::orb->string_to_object( (char*) (const char *) tmpstr);
      Observer_ptr obsobj=
	Observer::_narrow( (CORBA::Object_ptr) tmpobj);
      if ( CORBA::is_nil( obsobj)) continue;

      private_attach_add_observer( obsobj);
      CORBA::release( obsobj);
    }
    catch (...) {
      continue;
    }
  }
}

template <class Observer_Types>
void Observed_I<Observer_Types>::gc_worker() {
  RWTValDlist<Observer_ptr> dead_observers;
  RWTValDlist<Observer_ptr> c_observers;

  {
    auto_lock local_lock(*this);

    Basic_Observed_I<Observer_Types>::gc_worker();

    for (Observer_List_Iterator iterator(my_add_observer_list);
	 iterator(); )
      c_observers.insert( Observer::_duplicate( iterator.key()));
  }

  // Browse through the hash table
  for (RWTValDlistIterator<Observer_ptr> iterator( c_observers); 
       iterator(); ) {

    int dead= 0;
    try {
      dead= iterator.key()->_non_existent();
    }
    catch (...) {
      dead= 1;
    }
    if (dead) dead_observers.insert( iterator.key());
  }

  {
    auto_lock local_lock(*this);

    for (RWTValDlistIterator<Observer_ptr> diter( dead_observers);
	 diter(); ) {
      try {
	private_detach_add_observer( diter.key());
      } catch (...) {}
      try {
	private_detach_observer( diter.key(),
				 get_item_ids_list( diter.key()));
      } catch(...) {}
    }
  }
  for (RWTValDlistIterator<Observer_ptr> deliter( c_observers);
       deliter(); )
    CORBA::release( deliter.key());
}


