//
//  File:         badge_holder_rep_i.cc
//  Author:       Andrzej Uszok (uszok@ics.agh.edu.pl)
//  Date Created: 21.12.97
//
// General Description:
//   Implementation file for the Badge_Holder_Rep_I class
//

#ifdef ORBIX
#define EXCEPTIONS
#endif

#include "badge_holder_rep_i.h"
#include "common.h"
#include "Naming.hh"
#include "db_utils.hh"

//
// 		Constructor
//              ___________

   Badge_Holder_Rep_I::Badge_Holder_Rep_I (
#ifdef ObjectStore
    	os_database* db_p
#elif defined FilePersistency
    	const char *repository_name_p
#endif
    ) : Basic_Rep_I<Holder_Types, H_SDO_Types>(hash_badge_holder,
#ifdef ObjectStore
    		db_p
#elif defined FilePersistency
    		repository_name_p
#endif
        ) ,
	  repos_cond( & repos_guard.mutex()),
	  last_repos_resolved( 0)
{
  _instance = this;
  create_corba_objects();

#ifdef ORBIX
  observer_corba_object = (Domain_Repository::Static_Data_Observer_ptr) new 
   TIE_Domain_Repository_Static_Data_Observer(Badge_Holder_Rep_I)
        (this);
#elif defined OMNIORB
  observer_corba_object = 
    new TIE_CLASS( Domain_Repository,Static_Data_Observer)
      <Badge_Holder_Rep_I,0>(this);
  CONNECT_IMPL( Domain_Repository::Static_Data_Observer_var, 
		observer_corba_object);  
#endif
  
}

// this routine delays registering the reference of badge holder repository
// until all dependent servers are up or a timeout (1 minute) expires
//
void Badge_Holder_Rep_I::register_my_servers( void *args) {
  Badge_Holder_Rep_I &view= *((Badge_Holder_Rep_I*) args);

  {
    ABng_S_Guarded::auto_lock local_lock(view.repos_guard);
    
    while ( CORBA::is_nil( (ABng_Core_Management::Badge_Rep_ptr) 
			   view.r_badge_rep) ||
	    CORBA::is_nil( (ABng_Core_Management::Sensor_Rep_ptr)
			   view.r_sensor_rep) ||
	    CORBA::is_nil( (Domain_Repository::User_Rep_ptr)
			   view.r_user_rep) ||
	    CORBA::is_nil( (Domain_Repository::Equipment_Rep_ptr)
			   view.r_equipment_rep)) {

      unsigned long sec,nsec;
      int wait_ret;
     
      omni_thread::get_time( 	&sec, &nsec, 
				60, 0);
#ifdef OMNIORB
      wait_ret= view.repos_cond.timedwait( sec, nsec);
#else     
      wait_ret= view.repos_cond.timed_wait( sec, nsec);
#endif
      omni_thread::get_time( 	&sec, &nsec, 
				0, 0);
      if ( sec-view.last_repos_resolved > 60) // no repository resolved in last minute
	break;  // give it up, register your reference(s)
    }
  }
#ifdef DEBUG
  cerr << "Registering badge holder repository reference" << endl;
#endif
  view.load_observers();
  Bootstrap::register_server_in_NS( Bootstrap::SRV_Badge_Holder_Repository(), 
				    view.my_rep);
}

void 
   Badge_Holder_Rep_I::init (
	int cold_start_p
   ) 
{
  Basic_Rep_I<Holder_Types, H_SDO_Types>::init(cold_start_p);

  try {
    (new omni_thread( register_my_servers, this))->start();
    register_callback(Bootstrap::SRV_Badge_Repository());
    register_callback(Bootstrap::SRV_Sensor_Repository());
    register_callback(Bootstrap::SRV_User_Repository());
    register_callback(Bootstrap::SRV_Equipment_Repository());
    ps_dev= new ProcessControlImpl( "badge_holder_repository", this);
  }
  catch (CORBA::Exception &ex) {
    cerr << "Caught exception while contacting server manager" << ex << endl;
    exit(1);
  }
}

void 
   Badge_Holder_Rep_I::init_corba_object ( 
	Badge_Holder_I *item_p
   ) 
{
	 // create the TIE (CORBA) object for the C++ object
#ifdef ORBIX
#ifdef ObjectStore
	ObjectStoreTIE_ABng_Core_System_Badge_Holder(Badge_Holder_I)* new_badge_holder_tie =
		new(my_db, ObjectStoreTIE_ABng_Core_System_Badge_Holder(Badge_Holder_I)::get_os_typespec())
		ObjectStoreTIE_ABng_Core_System_Badge_Holder(Badge_Holder_I) (item_p);
#else
	TIE_ABng_Core_System_Badge_Holder(Badge_Holder_I)* new_badge_holder_tie = 
		new TIE_ABng_Core_System_Badge_Holder(Badge_Holder_I) (item_p);
#endif
#elif defined OMNIORB
  ABng_Core_System::Badge_Holder_ptr new_badge_holder_tie=
    new TIE_CLASS(ABng_Core_System,Badge_Holder)<Badge_Holder_I,0>(item_p);
  CONNECT_IMPL( ABng_Core_System::Badge_Holder_var, new_badge_holder_tie);     
#endif
	item_p->set_corba_object(new_badge_holder_tie);
}

void 
   Badge_Holder_Rep_I::create_corba_objects (
   ) 
{
#ifdef ORBIX
#ifdef FilePersistency
	// create a TIE for the object
	my_rep = new TIE_ABng_Core_Management_Badge_Holder_Rep(Badge_Holder_Rep_I) (this);
#endif /* FilePersistency */
#elif defined OMNIORB
  my_rep =
    new TIE_CLASS(ABng_Core_Management,Badge_Holder_Rep)<Badge_Holder_Rep_I,0>(this);
  CONNECT_IMPL( ABng_Core_System::Badge_Holder_Rep_var,my_rep);	
#endif
}

Badge_Holder_Rep_I* Badge_Holder_Rep_I::_instance = 0;

Badge_Holder_Rep_I* 
   Badge_Holder_Rep_I::instance(
   )
{
   	return _instance;
}
//
// 		Destructor
//		__________

   Badge_Holder_Rep_I::~Badge_Holder_Rep_I()
{
  	destroy_corba_objects();
	delete ps_dev;
}

void 
   Badge_Holder_Rep_I::destroy_corba_objects (
   ) 
{ 
	CORBA::release(my_rep);
}

//
// Private functions
//

void 
   Badge_Holder_Rep_I::reference_resolved ( 
	const ABng_Server_Manager::Name &name,
	CORBA::Object_var server_ref,
	ABng_Server_Manager::uid callback_id
   ) 
{
#ifdef DEBUG
  cerr << "<<1 Resolved " << name << endl;
#endif
  auto_lock local_lock(*this);
#ifdef DEBUG
  cerr << "<<2 Resolved (and locked) " << name << endl;
#endif

  if (name == Bootstrap::SRV_Badge_Repository()) {
    cerr << "Badge_rep" << endl;
    try {
      r_badge_rep = 
	ABng_Core_Management::Badge_Rep::_narrow(server_ref);
      cerr << "narrowed" << endl;
      if (CORBA::is_nil(r_badge_rep))
	cerr << "is_nil" << endl;
      resolve_badges();
      cerr << "resolved" << endl;
    }
    catch (CORBA::Exception &ex) {
      cerr << ex << endl;
      r_badge_rep = ABng_Core_Management::Badge_Rep::_nil();
    }
    catch (...) {
      cerr << "Exception" << endl;
      r_badge_rep = ABng_Core_Management::Badge_Rep::_nil();
    }
  }
  else if (name == Bootstrap::SRV_Sensor_Repository()) {
    try {
      r_sensor_rep = 
	ABng_Core_Management::Sensor_Rep::_narrow(server_ref);
      resolve_home_sensors();
    }
    catch (...) {
      r_sensor_rep = ABng_Core_Management::Sensor_Rep::_nil();
    }
  }
  else if (name == Bootstrap::SRV_User_Repository()) {
    try {
      r_user_rep = 
	Domain_Repository::User_Rep::_narrow(server_ref);
      resolve_users();
    }
    catch (...) {
      r_user_rep = Domain_Repository::User_Rep::_nil();
    }
  }
  else if (name == Bootstrap::SRV_Equipment_Repository()) {
    try {
      r_equipment_rep = 
	Domain_Repository::Equipment_Rep::_narrow(server_ref);
      resolve_equipments();
    }
    catch (...) {
      r_equipment_rep = Domain_Repository::Equipment_Rep::_nil();
    }
  }
  omni_thread::get_time( &last_repos_resolved, &last_repos_resolved_us, 
			 0, 0);
  repos_cond.signal();

#ifdef DEBUG
  cerr << ">> Resolved (about to unlock)" << name << endl;
#endif
}

void 
   Badge_Holder_Rep_I::resolve_badges(
   ) 
{
  if (r_badge_rep->_non_existent()) 
    throw 0;
  
  RWTPtrHashTableIterator<Badge_Holder_I> iterator(my_item_list);
  while (iterator()) {
    iterator.key()->resolve_my_badge(r_badge_rep);
    iterator.key()->resolve_current_badge(r_badge_rep);
    
    Observer_List *observer_list_p = 
      get_item_observer_list(iterator.key()->item_id());
    notify_observers( *observer_list_p, 
		      iterator.key(), 
		      Domain_Repository::OBJECT_UPDATED);
  }
}

void 
   Badge_Holder_Rep_I::resolve_home_sensors(
   ) 
{
  if (r_sensor_rep->_non_existent()) 
    throw 0;

  RWTPtrHashTableIterator<Badge_Holder_I> iterator(my_item_list);
  while (iterator()) {
    iterator.key()->resolve_home_sensor(r_sensor_rep);
    
    Observer_List *observer_list_p = 
      get_item_observer_list(iterator.key()->item_id());
    notify_observers( *observer_list_p, 
		      iterator.key(), 
		      Domain_Repository::OBJECT_UPDATED);
  }
}

void 
   Badge_Holder_Rep_I::resolve_users(
   ) 
{
  if (r_user_rep->_non_existent()) 
    throw 0;

  ABng_Core_System::Badge_Holder::Pattern pattern;
  pattern.is_any = 1;
  ABng_Core_System::Badge_Holders_var descrs = private_find(pattern);

  for(int i = 0; i < descrs->length(); i ++) {
    Item_I *item = deref_item(descrs[i]);
    
    if (item->get_descr().type_specific_object._d() !=
	ABng_Core::LOCATABLE_USER) continue;

    item->resolve_user(r_user_rep);

    if (CORBA::is_nil(item->get_descr().type_specific_object.user_obj()))
      private_remove( descrs[i]);
    else {
      Observer_List *observer_list_p = 
	get_item_observer_list( item->item_id());
      notify_observers( *observer_list_p, 
			item, 
			Domain_Repository::OBJECT_UPDATED);
    }
  }
}

void 
   Badge_Holder_Rep_I::resolve_equipments(
   ) 
{
  if (r_equipment_rep->_non_existent()) 
    throw 0;

  RWTPtrHashTableIterator<Badge_Holder_I> iterator(my_item_list);
  while (iterator()) 
    if (iterator.key()->get_descr().type_specific_object._d() == ABng_Core::LOCATABLE_EQUIPMENT) {
      iterator.key()->resolve_equipment(r_equipment_rep);
      
      Observer_List *observer_list_p = 
	get_item_observer_list(iterator.key()->item_id());
      notify_observers( *observer_list_p, 
			iterator.key(), 
			Domain_Repository::OBJECT_UPDATED);
    }
}

void Badge_Holder_Rep_I::
notify ( Domain_Repository::Repository_Item_Id id,
	 const Domain_Repository::Static_Data_Observer::Static_Data_Object& sd, 
	 Domain_Repository::Change_Type change
	 ENV_DEF(env)) {

  auto_lock local_lock(*this);

  switch(change) {
  case Domain_Repository::OBJECT_ADDED:
    break;
    
  case Domain_Repository::OBJECT_REMOVED: {
#ifdef DEBUG
    cerr << "User object removed " << (unsigned) id << endl;	
#endif
    ABng_Core_System::Badge_Holder::Pattern pattern;
    Domain_Repository::User::Pattern upattern;
    upattern.is_any= 0;
    upattern.item_id_pattern.is_any= 0;
    upattern.item_id_pattern.value= id;
    upattern.uid_pattern.is_any=
      upattern.forename_pattern.is_any= 
      upattern.surname_pattern.is_any=
      upattern.position_pattern.is_any=
      upattern.email_pattern.is_any=
      upattern.phone_pattern.is_any= 1;
    
    pattern.is_any = 0;
    pattern.item_id_pattern.is_any= 1;
    pattern.my_badge_pattern.is_any= 1;
    pattern.current_badge_pattern.is_any= 1;
    pattern.home_sensor_pattern.is_any= 1;
    pattern.type_specific_object_pattern.is_any= 0;
    pattern.type_specific_object_pattern.object_pattern.user_obj_pattern( upattern);
    pattern.assigned_pattern.is_any= 
      pattern.visitor_pattern.is_any= 1;

    ABng_Core_System::Badge_Holders_var bholders;
    try {
      bholders = private_find(pattern);
    } catch(CORBA::Exception &ex) {
      cerr << ex << endl;
      return;
    } catch(...) {
      return;
    }
    
    for(int i = 0; i < bholders->length(); i ++) {
      try {
	private_remove(bholders[i]);
      } catch(...) {
      }
    }
    
    break;
  }
  case Domain_Repository::OBJECT_UPDATED:
    break;
  }
}

void Badge_Holder_Rep_I::abort ( ENV_DEF1(env)) {
#ifdef DEBUG
  cerr << "Badge_Holder_Rep_I: abort called!!!" << endl;
#endif
}



DB_Header* 
   Badge_Holder_Rep_I::create_db_header (
   ) 
{
  return new DB_Header("Badge_Holder Repository", "1.0", "\
#       <data_record>::=  	<item_id> <my_badge_id> <current_badge_id> <home_sensor_id>\n\
#       			<badge_holder_kind> <type_specific_object_id>\n\
#                               <is_assigned> <is_visitor>\n\
#\n\
#	<my_badge_id>::=		<item_id>\n\
#	<current_badge_id>::=		<item_id>\n\
#	<home_sensor_id>::=		<item_id>\n\
#	<type_specific_object_id>::=	<item_id>\n\
#	<badge_holder_kind>::=		LOCATABLE_USER (0)\n\
#	<badge_holder_kind>::=		LOCATABLE_EQUIPMENT (1)\n\
#\n\
");
}

unsigned 
   Badge_Holder_Rep_I::hash_badge_holder(
	const Badge_Holder_I& badge_holder_i_p
   )
{
	return badge_holder_i_p.get_current_badge_id();
}

void
   Badge_Holder_Rep_I::gather_badge_holder(
	Badge_Holder_I *current_badge_holder,
	void* apply_data
   )
{
	// project apply_data (void*) to Badge_Holders*
	ABng_Core_System::Badge_Holders *collected_badge_holders =
		(ABng_Core_System::Badge_Holders*)apply_data;

	// increase the length of the container
	collected_badge_holders->length(collected_badge_holders->length() + 1);

	// put duplicated badge_holders (CORBA object) into the container
	(*collected_badge_holders)[collected_badge_holders->length() - 1] =
		ABng_Core_System::Badge_Holder::_duplicate(current_badge_holder->get_corba_object());

}

    Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct::MatchBadge_HolderApplyDataStruct(
	ABng_Core_System::Badge_Holders* _found_badge_holders_p,
   	const ABng_Core_System::Badge_Holder::Pattern& _data_record_pattern_p,
        ABng_Core_Management::Badge_Rep_var _badge_rep,
        ABng_Core_Management::Sensor_Rep_var _sensor_rep,
        Domain_Repository::User_Rep_var _user_rep,
        Domain_Repository::Equipment_Rep_var _equipment_rep
    ):
	 	_found_badge_holders(_found_badge_holders_p),
		_data_record_pattern(_data_record_pattern_p),
		_my_badge_descrs(0),
		_current_badge_descrs(0),
		_home_sensor_descrs(0),
		_user_descrs(0),
		_equipment_descrs(0),
		_r_badge_rep(_badge_rep),
		_r_sensor_rep(_sensor_rep),
		_r_user_rep(_user_rep),
		_r_equipment_rep(_equipment_rep) 
{

}

Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct::~MatchBadge_HolderApplyDataStruct() 
{
  if(_my_badge_descrs) delete _my_badge_descrs;
  if(_current_badge_descrs) delete _current_badge_descrs;
  
  if(_home_sensor_descrs) delete _home_sensor_descrs;
  
  if(_user_descrs) delete _user_descrs;
  if(_equipment_descrs) delete _equipment_descrs;
}


int
   Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct::match_my_badge(
	const ABng_Core_System::Badge::Pattern& pattern,
	Domain_Repository::Repository_Item_Id _id
     )
{
  if(!_my_badge_descrs) {
    if(CORBA::is_nil((ABng_Core_Management::Badge_Rep_ptr)_r_badge_rep))  
      throw Domain_Repository::Destination_Unreachable("badge_repository");

    try {
      _my_badge_descrs = _r_badge_rep->find_values(pattern);  
    } catch(...) {
      throw Domain_Repository::Destination_Unreachable("badge_repository");
    }
  }

  for(int i = 0; i < _my_badge_descrs->length(); i ++)
    if((*_my_badge_descrs)[i].item_id == _id) 
      return 1;
  return 0;
}

int
   Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct::match_current_badge(
	const ABng_Core_System::Badge::Pattern& pattern,
	Domain_Repository::Repository_Item_Id _id
     )
{
  if(!_current_badge_descrs) {
    if(CORBA::is_nil((ABng_Core_Management::Badge_Rep_ptr)_r_badge_rep)) 
      throw Domain_Repository::Destination_Unreachable("badge_repository");
    
    try {
      _current_badge_descrs = _r_badge_rep->find_values(pattern);
      
    } catch(...) {
      throw Domain_Repository::Destination_Unreachable("badge_repository");
    }
  }
  
  for(int i = 0; i < _current_badge_descrs->length(); i ++)
    if((*_current_badge_descrs)[i].item_id == _id) 
      return 1;
  return 0;
}


int
   Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct::match_home_sensor(
	const ABng_Core_System::Sensor::Pattern& pattern,
	Domain_Repository::Repository_Item_Id _id
     )
{
  if(!_home_sensor_descrs) {
    if(CORBA::is_nil((ABng_Core_Management::Sensor_Rep_ptr)_r_sensor_rep)) 
      throw Domain_Repository::Destination_Unreachable("sensor_repository");
    
    try {
      _home_sensor_descrs = _r_sensor_rep->find_values(pattern);

    } catch(...) {
      throw Domain_Repository::Destination_Unreachable("sensor_repository");
    }
  }
  
  for(int i = 0; i < _home_sensor_descrs->length(); i ++)
    if((*_home_sensor_descrs)[i].item_id == _id) 
      return 1;
  return 0;
}

int
   Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct::match_user(
	const Domain_Repository::User::Pattern& pattern,
	Domain_Repository::Repository_Item_Id _id
     )
{
  if(!_user_descrs) {
    if(CORBA::is_nil((Domain_Repository::User_Rep_ptr)_r_user_rep)) 
      throw Domain_Repository::Destination_Unreachable("user_repository");
    
    try {
      _user_descrs = _r_user_rep->find_values(pattern);
      
    } catch(...) {
      throw Domain_Repository::Destination_Unreachable("user_repository");
    }
  }

  for(int i = 0; i < _user_descrs->length(); i ++)
    if((*_user_descrs)[i].item_id == _id) 
      return 1;
  return 0;
}

int
   Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct::match_equipment(
	const Domain_Repository::Equipment::Pattern& pattern,
	Domain_Repository::Repository_Item_Id _id
     )
{
  if(!_equipment_descrs) {
    if(CORBA::is_nil((Domain_Repository::Equipment_Rep_ptr)_r_equipment_rep)) 
      throw Domain_Repository::Destination_Unreachable("equipment_repository");
    
    try {
      _equipment_descrs = _r_equipment_rep->find_values(pattern);
      
    } catch(...) {
      throw Domain_Repository::Destination_Unreachable("equipment_repository");
    }
  }
  
  for(int i = 0; i < _equipment_descrs->length(); i ++)
    if((*_equipment_descrs)[i].item_id == _id) 
      return 1;
  return 0;
}

void
   Badge_Holder_Rep_I::match_badge_holder(
	Badge_Holder_I *current_badge_holder,
	void* apply_data
   )
{

  struct Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct* match_struct =
    (struct Badge_Holder_Rep_I::MatchBadge_HolderApplyDataStruct*)apply_data;

  const ABng_Core_System::Badge_Holder::Pattern& data_record_pattern = 
    match_struct->_data_record_pattern;

  // check general patten matching rule
  if (data_record_pattern.is_any){
    // any badge_holder matches this pattern
    gather_badge_holder(current_badge_holder, match_struct->_found_badge_holders);
    return;
  }

  // check matching of item_id
  if (!data_record_pattern.item_id_pattern.is_any &&
      (data_record_pattern.item_id_pattern.value != 
       current_badge_holder->get_descr().item_id))
    // this badge_holder item_id does not match pattern 
    return;
  
  // check matching of my_badge
  if (!data_record_pattern.my_badge_pattern.is_any) {
    // check matching of my_badge item_id
    if (!data_record_pattern.my_badge_pattern.item_id_pattern.is_any &&
	data_record_pattern.my_badge_pattern.item_id_pattern.value !=
	current_badge_holder->get_my_badge_id())
      // my_badge item_id does not match pattern 
      return;
    
    // check matching of other data
    if (!data_record_pattern.my_badge_pattern.id_pattern.is_any) {
      if (!match_struct->match_my_badge(data_record_pattern.my_badge_pattern,
					current_badge_holder->get_my_badge_id()))
	return;
    }
  }

  // check matching of current_badge
  if (!data_record_pattern.current_badge_pattern.is_any) {
    // check matching of current_badge item_id
    if (!data_record_pattern.current_badge_pattern.item_id_pattern.is_any &&
	data_record_pattern.current_badge_pattern.item_id_pattern.value !=
	current_badge_holder->get_current_badge_id())
      // current_badge item_id does not match pattern 
      return;
    
    // check matching of other data
    if (!data_record_pattern.current_badge_pattern.id_pattern.is_any) {
      if (!match_struct->match_current_badge(data_record_pattern.current_badge_pattern,
					     current_badge_holder->get_current_badge_id()))
	return;
    }
  }
  
  // check matching of home_sensor
  if (!data_record_pattern.home_sensor_pattern.is_any) {
    // check matching of home_sensor item_id
    if (!data_record_pattern.home_sensor_pattern.item_id_pattern.is_any &&
	data_record_pattern.home_sensor_pattern.item_id_pattern.value !=
	current_badge_holder->get_home_sensor_id())
      // home_sensor item_id does not match pattern 
      return;
    
    // check matching of other data
    if (!data_record_pattern.home_sensor_pattern.id_pattern.is_any) {
      if (!match_struct->match_home_sensor(data_record_pattern.home_sensor_pattern,
					   current_badge_holder->get_home_sensor_id()))
	return;
    }
  }

  // check matching of type_specific_object
  if (!data_record_pattern.type_specific_object_pattern.is_any) {
    ABng_Core::Locatable_Type_Specific_Object_Pattern::Ref_Pattern _object_pattern =
      data_record_pattern.type_specific_object_pattern.object_pattern;
    ABng_Core::Locatable_Type_Specific_Object _type_specific_object = 
      current_badge_holder->get_descr().type_specific_object;
    
    // check matching of type_specific_object kind (user or equipment)
    if (_object_pattern._d() != _type_specific_object._d())
      // this badge holder kind does not match kind of pattern
      return;
    
    if (_object_pattern._d() == ABng_Core::LOCATABLE_USER) { 
      
      Domain_Repository::User::Pattern user_pattern = _object_pattern.user_obj_pattern();
			
      // check matching of user item_id
      if (!user_pattern.item_id_pattern.is_any && user_pattern.item_id_pattern.value !=
	  current_badge_holder->get_type_specific_object_id())
	// user item_id does not match pattern 
	return;
      
      // check matching of other data
      if (!user_pattern.uid_pattern.is_any || !user_pattern.forename_pattern.is_any 
	  || !user_pattern.surname_pattern.is_any || !user_pattern.position_pattern.is_any
	  || !user_pattern.email_pattern.is_any || !user_pattern.phone_pattern.is_any) {
	if (!match_struct->match_user(user_pattern, current_badge_holder->get_type_specific_object_id()))
	  // this user id was not found in ids set which match pattern
	  return;
      }
    }
    else {
      Domain_Repository::Equipment::Pattern equipment_pattern = _object_pattern.equipment_obj_pattern();
      
      // check matching of equipment item_id
      if (!equipment_pattern.item_id_pattern.is_any && equipment_pattern.item_id_pattern.value !=
	  current_badge_holder->get_type_specific_object_id())
	// equipment item_id does not match pattern 
	return;
      
      // check matching of other data
      if (!equipment_pattern.meta_data_pattern.is_any || !equipment_pattern.name_pattern.is_any 
	  || !equipment_pattern.last_command_pattern.is_any || !equipment_pattern.property_list_pattern.is_any) {
	if (!match_struct->match_equipment(equipment_pattern,
			   	   current_badge_holder->get_type_specific_object_id()))
	  // this equipment id was not found in ids set which match pattern
	  return;
      }
    }
  }
  
  if (!data_record_pattern.assigned_pattern.is_any)
    if (data_record_pattern.assigned_pattern.value !=
	current_badge_holder->get_descr().assigned) return;
  
  if (!data_record_pattern.visitor_pattern.is_any)
    if (data_record_pattern.visitor_pattern.value !=
	current_badge_holder->get_descr().visitor) return;
  
  // this badge_holder matches the pattern
  gather_badge_holder(current_badge_holder, match_struct->_found_badge_holders);
}

int 
   Badge_Holder_Rep_I::validate_add_data (
	const ABng_Core_System::Badge_Holder::Description &data_record
   ) 
{
#ifdef ORBIX
  	// a very nasty thing (to cast away a const), 
  	// but masks Orbix bug (ver. 2.2, 2.3)
  	if (CORBA::is_nil(data_record.my_badge)) 
    		((ABng_Core_System::Badge_Holder::Description&) data_record).my_badge = NULL;
  	if (CORBA::is_nil(data_record.current_badge)) 
    		((ABng_Core_System::Badge_Holder::Description&) data_record).current_badge = NULL;
  	if (CORBA::is_nil(data_record.home_sensor)) 
    		((ABng_Core_System::Badge_Holder::Description&) data_record).home_sensor = NULL;
	if (((ABng_Core_System::Badge_Holder::Description&) data_record).type_specific_object._d() 
	      == ABng_Core::LOCATABLE_USER) {
		if (CORBA::is_nil(data_record.type_specific_object.user_obj())) 
    			((ABng_Core_System::Badge_Holder::Description&)
				 data_record).type_specific_object.user_obj(NULL);

	} else {
  		if (CORBA::is_nil(data_record.type_specific_object.equipment_obj())) 
    			((ABng_Core_System::Badge_Holder::Description&)
				 data_record).type_specific_object.equipment_obj(NULL);
        } 
#endif
	return 1;
}

int 
   Badge_Holder_Rep_I::validate_update_data (
	ABng_Core_System::Badge_Holder_ptr item_p,
	const ABng_Core_System::Badge_Holder::Description& data_record
   ) 
{
#ifdef ORBIX
   	// a very nasty thing (to cast away a const), 
  	// but masks Orbix bug (ver. 2.2, 2.3)
  	if (CORBA::is_nil(data_record.my_badge)) 
    		((ABng_Core_System::Badge_Holder::Description&) data_record).my_badge = NULL;
  	if (CORBA::is_nil(data_record.current_badge)) 
    		((ABng_Core_System::Badge_Holder::Description&) data_record).current_badge = NULL;
  	if (CORBA::is_nil(data_record.home_sensor)) 
    		((ABng_Core_System::Badge_Holder::Description&) data_record).home_sensor = NULL;
	if (((ABng_Core_System::Badge_Holder::Description&) data_record).type_specific_object._d() 
	      == ABng_Core::LOCATABLE_USER) {
		if (CORBA::is_nil(data_record.type_specific_object.user_obj())) 
    			((ABng_Core_System::Badge_Holder::Description&)
				 data_record).type_specific_object.user_obj(NULL);

	} else {
  		if (CORBA::is_nil(data_record.type_specific_object.equipment_obj())) 
    			((ABng_Core_System::Badge_Holder::Description&)
				 data_record).type_specific_object.equipment_obj(NULL);
        } 
#endif
	return 1;
}

void 
   Badge_Holder_Rep_I::construct_data_object (
	ABng_Core_System::System_Static_Data_Observer::System_Static_Data_Object& data_object_p,
	ABng_Core_System::Badge_Holder_ptr item_p
   ) 
{
	// construct union to store reference of added entity
	data_object_p.badge_holder_obj(ABng_Core_System::Badge_Holder::_duplicate(item_p));
}

void 
   Badge_Holder_Rep_I::collect_values ( 
	Badge_Holder_I *lt, 
	ABng_Core_System::Badge_Holder::Value_Description& ld
   ) 
{
	ld = *(lt->private_value_descr());
}

ABng_Core_System::Badge_Holder::Description 
  Badge_Holder_Rep_I::get_null_description( Item_I *item) {
  
  Description d;
  if (item) d.item_id= item->item_id();
  if (item)
    if (item->kind() == ABng_Core::LOCATABLE_EQUIPMENT) {
      d.type_specific_object.equipment_obj( Domain_Repository::Equipment::_nil());
      return d;
    }
  d.type_specific_object.user_obj( Domain_Repository::User::_nil());
  return d;
}

ABng_Core_System::Badge_Holder::Value_Description 
  Badge_Holder_Rep_I::get_null_value_description( Item_I *item) {

  Value_Description d;
  if (item) d.item_id= item->item_id();
  if (item) 
    if (item->kind() == ABng_Core::LOCATABLE_USER) {
      d.type_specific_object.equipment_description( 
				  Domain_Repository::Equipment::Value_Description());
      return d;
    }
  d.type_specific_object.user_description( 
				  Domain_Repository::User::Value_Description());
  return d;
}



//
// Implementation of IDL operations of the System_Static_Data_Observed_R interface
//                                         _____________________________

ABng_Core_System::System_Static_Data_Observed_R::Observer_Entries*
   Badge_Holder_Rep_I::system_static_data_observer_list (
	ENV_DEF1(env)
   ) 
		throw (CORBA::SystemException)
{	
	return observer_list(USE_ENV1(env));
}

ABng_Core_System::System_Static_Data_Observed_E::System_Static_Data_Observers*
   Badge_Holder_Rep_I::add_system_static_data_observer_list (
	ENV_DEF1(env)
   ) 
		throw (CORBA::SystemException)
{
	return add_observer_list(USE_ENV1(env));
}
                        
void 
   Badge_Holder_Rep_I::attach_system_static_data_observer (
	ABng_Core_System::System_Static_Data_Observer_ptr obs,  
	const Domain_Repository::Repository_Item_Ids& observed_id_list
	ENV_DEF(env)
   )
		throw (CORBA::SystemException, 
                       Domain_Repository::Unknown_Repository_Item_Ids)
{
	attach_observer(obs, observed_id_list USE_ENV(env));
}

void 
   Badge_Holder_Rep_I::detach_system_static_data_observer (
	ABng_Core_System::System_Static_Data_Observer_ptr obs, 
	const Domain_Repository::Repository_Item_Ids& observed_id_list
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException, 
                       Domain_Repository::Unknown_Repository_Item_Ids, 
                       Domain_Repository::Unknown_Observer,
		       Domain_Repository::Unmatched_Observer_Item_Ids)
{
	detach_observer(obs, observed_id_list USE_ENV(env));
}

void 
   Badge_Holder_Rep_I::attach_add_system_static_data_observer (
	ABng_Core_System::System_Static_Data_Observer_ptr obs  
	ENV_DEF(env)
   )
		throw (CORBA::SystemException)
{
	attach_add_observer(obs USE_ENV(env));
}

void 
   Badge_Holder_Rep_I::detach_add_system_static_data_observer (
	ABng_Core_System::System_Static_Data_Observer_ptr obs 
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException, 
                       Domain_Repository::Unknown_Observer)
{
	detach_add_observer(obs USE_ENV(env));
}


// Suplementary operation

ABng_Core_System::System_Static_Data_Observed_E::System_Static_Data_Observers*
   Badge_Holder_Rep_I::system_static_data_observer_list (
	Domain_Repository::Repository_Item_Id item_id_p
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException)
{
	return get_item_observers(item_id_p USE_ENV(env));
}

//
// Implementation of IDL operations of the Badge_Holder_Rep interface
//                                         ________________

ABng_Core_System::Badge_Holders* 
   Badge_Holder_Rep_I::badge_holder_list (
	ENV_DEF1(env)
   ) 
		throw (CORBA::SystemException)
{
	return item_list(USE_ENV1(env));
}
                
ABng_Core_System::Badge_Holder_ptr 
   Badge_Holder_Rep_I::add (
	const ABng_Core_System::Badge_Holder::Description& data_record
	ENV_DEF(env)
   )
		throw (CORBA::SystemException, 
                       Domain_Repository::Duplicate_Data_Record)
{
	return Basic_Rep_I<Holder_Types, H_SDO_Types>::add(data_record USE_ENV(env));
}

void 
   Badge_Holder_Rep_I::remove (
	ABng_Core_System::Badge_Holder_ptr object_to_remove 
	ENV_DEF(env)
   )
		throw (CORBA::SystemException, 
                       Domain_Repository::Unknown_Object_Ref)
{
	Basic_Rep_I<Holder_Types, H_SDO_Types>::remove(object_to_remove USE_ENV(env));
}

ABng_Core_System::Badge_Holders* 
   Badge_Holder_Rep_I::find (
	const ABng_Core_System::Badge_Holder::Pattern& data_record_pattern
	ENV_DEF(env)
   )
		throw (CORBA::SystemException,
                       Domain_Repository::Search_Not_Implemented)
{
  	auto_lock local_lock(*this);

  	return private_find(data_record_pattern USE_ENV(env));
}

ABng_Core_System::Badge_Holders* 
   Badge_Holder_Rep_I::private_find (
	const ABng_Core_System::Badge_Holder::Pattern& data_record_pattern
	ENV_DEF(env)
   )
		throw (CORBA::SystemException,
                       Domain_Repository::Search_Not_Implemented)
{
  // create a structure consisting of a pointer to the container
  // a reference of the pattern and references to related repositories
  MatchBadge_HolderApplyDataStruct apply_data(
		(ABng_Core_System::Badge_Holders*)new 
		ABng_Core_System::Badge_Holders(my_item_list.entries()),
		data_record_pattern, r_badge_rep, r_sensor_rep, r_user_rep, r_equipment_rep);

  // browse through hash table to collect badge_holders
  my_item_list.apply(match_badge_holder, (void*)&apply_data);
  
  cerr << "private_find" << endl;
  // pointer to found_badge_holders was stored inside apply_data thus return it
  return apply_data._found_badge_holders;
}

ABng_Core_System::Badge_Holder::Value_Descriptions* 
   Badge_Holder_Rep_I::find_values (
	const ABng_Core_System::Badge_Holder::Pattern& data_record_pattern
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException,
		       Domain_Repository::Destination_Unreachable,
                       Domain_Repository::Search_Not_Implemented)
{
  return Basic_Rep_I<Holder_Types, H_SDO_Types>::find_values(data_record_pattern USE_ENV(env));
}

void 
   Badge_Holder_Rep_I::update (
	ABng_Core_System::Badge_Holder_ptr object_to_update, 
	const ABng_Core_System::Badge_Holder::Description& data_record_pattern
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException, 
		       Domain_Repository::Unknown_Object_Ref,
		       Domain_Repository::Duplicate_Data_Record)
{
	Basic_Rep_I<Holder_Types, H_SDO_Types>::update(object_to_update, data_record_pattern USE_ENV(env));
}









