// $Id: sched_barrier.cc,v 1.4 1998/07/29 14:00:43 mencnar Exp $
//
// $Source: /home/iris/grant/abng/CVS/sources/servers/scheduler/sched_barrier.cc,v $
//
// $Log: sched_barrier.cc,v $
// Revision 1.4  1998/07/29 14:00:43  mencnar
// bug fixed:
//  an answer (from badge) _was_ expected only from a sensor that
//  sent a request,
//
// Revision 1.3  1998/06/18 10:21:56  mencnar
// moved to CVS
//
// Revision 1.2  1997/11/03 09:35:52  mencnar
// Added destructor for Sched_Barrier
//
// Revision 1.1  1997/10/23 09:16:19  mencnar
// Initial revision
//
//


// Temporary log:
// Fixed potential memory leak ( handling replies)
//

#include "sched_main.hh"
#include <iostream.h>


//---------------------------------------------------------------Sched_Barrier

Sched_Barrier::~Sched_Barrier() {
#ifdef DEBUG
  cerr << "Removing barrier ( " << sighting_queue.entries() << " request in queue)" << endl;
#endif
  if (my_reply) delete my_reply;
}

void Sched_Barrier::insert_request( Sched_Request &req) {
  auto_lock local_lock( *this);

  sighting_queue.append( &req);
}

void Sched_Barrier::remove_request( Sched_Request &req) {
  auto_lock local_lock( *this);

  sighting_queue.remove( &req);
  if (!barrier_id.is_badge_id /* sensor command*/ ) {
    if (!sighting_queue.isEmpty())
      sighting_queue[0]->cond_sighting_arrived.signal();
  }
}


// private method : no need to lock
//
void Sched_Barrier::begin_request( Sched_Request &req) {
  req.last_pos= last_sighting_pos;
  req.last_sighting_time= last_sighting_time;
  request_in_progress= 1;
}

void Sched_Barrier::end_request() {
  auto_lock local_lock( *this);

  request_in_progress= 0;
}

void Sched_Barrier::wait_till( ABng_Core::Time &till,
			       Sched_Request &req) {
  auto_lock local_lock( *this);
#ifdef OMNIORB
  req.cond_sleep.timedwait( till.seconds, till.microseconds*1000);
#else
  req.cond_sleep.timed_wait( till.seconds, till.microseconds*1000);
#endif
}

//
// Returns:	-1	-if request was cancelled
//			 the controlling thread should exit immediately
//		 0	-on timeout
//		+1	-if sighting was acquired
//
// Note: if function returns +1, the caller is responsible to terminate
//       request processing with end_request(), it will allow
//       next thread waiting on this barrier to proceed
//
int Sched_Barrier::wait_for_sighting( 
	ABng_Core::Time &timeout,
	Sched_Request &req) {
  unsigned long sec,nsec;

  auto_lock local_lock( *this);
  // enter critical section

  int wait_ret;

  if (req.cancelled()) return -1;
    
  if (!barrier_id.is_badge_id /* sensor command*/ ) {
#ifdef OMNIORB
    wait_ret= 1;
#else
    wait_ret= ETIMEDOUT+1;
#endif
    if (request_in_progress) {
      omni_thread::get_time( 	&sec, &nsec, 
				timeout.seconds, timeout.microseconds*1000);
#ifdef OMNIORB
      wait_ret= req.cond_sighting_arrived.timedwait( sec, nsec);
#else     
      wait_ret= req.cond_sighting_arrived.timed_wait( sec, nsec);
#endif
    }
    if (req.cancelled()) return -1;
#ifdef OMNIORB
    if (wait_ret == 0) return 0;
#else
    if (wait_ret == ETIMEDOUT) return 0;
#endif
    begin_request( req);
    return 1;
  }

  if (!request_in_progress) {
    omni_thread::get_time( &sec, &nsec);
    ABng_Core::Time current_time;
    current_time.seconds= sec;
    current_time.microseconds= nsec/1000;

    ABng_Core::Time x_time;
    x_time.seconds= last_sighting_time.seconds+sighting_timeout.seconds;
    x_time.microseconds= last_sighting_time.microseconds+sighting_timeout.microseconds;

    if (Sched_Request::compare_time( current_time, x_time)< 0) {
      if (req.to_nearest_sensor) {
	begin_request( req);
	return +1;
      }

      for (size_t j=0; j<req.barrier_ids.entries(); j++)
	if (req.barrier_ids[j] == last_sighting_pos) {
	  begin_request( req);
	  return +1;
	}
    }
  }
 
  omni_thread::get_time( 	&sec, &nsec, 
				timeout.seconds, timeout.microseconds*1000);

#ifdef OMNIORB
  wait_ret= req.cond_sighting_arrived.timedwait( sec, nsec);
#else
  wait_ret= req.cond_sighting_arrived.timed_wait( sec, nsec);
#endif
  if (req.cancelled()) return -1;
#ifdef OMNIORB
  if (wait_ret == 0) return 0;
#else
  if (wait_ret == ETIMEDOUT) return 0;
#endif
  begin_request( req);
  return +1;
}

void Sched_Barrier::sighting_arrived(
	const Sched_Barrier_Id &new_loc) {
  auto_lock local_lock( *this);

  unsigned long sec,nsec;
  omni_thread::get_time( &sec, &nsec);

  last_sighting_time.seconds= sec;
  last_sighting_time.microseconds= nsec/1000;
  last_sighting_pos= new_loc;
  if (request_in_progress || sighting_queue.isEmpty()) {
#ifdef DEBUG
    if (request_in_progress) cout << "request in progress" << endl;
    if (sighting_queue.isEmpty()) cout << "sighting queue empty" << endl;
#endif
    return;
  }

  for (size_t i= 0; i<sighting_queue.entries(); i++) {
    Sched_Request &req= * sighting_queue[i];

    if (req.to_nearest_sensor) {
      req.cond_sighting_arrived.signal();
      break;
    }

    for (size_t j=0; j<req.barrier_ids.entries(); j++)
      if (req.barrier_ids[j] == new_loc) {
#ifdef DEBUG
	cout << "Waking up request" << endl;
#endif
	req.cond_sighting_arrived.signal(); 
					// wake up single
					// thread waiting for sighting
      	break;
      }
  }
}

//
// Returns:	-1	-if request was cancelled,
//			 the controlling thread should exit immediately
//		 0	-on timeout
//		+1	-if reply arrived
//
int Sched_Barrier::wait_for_reply(
	ABng_Core::Time &timeout,
	Sched_Request &req) {
  unsigned long sec, nsec;

  auto_lock local_lock( *this);

  if (my_reply) {
    Sched_Barrier_Id xlastpos( barrier_id.badge_id,
			       req.last_pos.network_id,
			       req.last_pos.station_id,
			       req.last_pos.group_name);
				 
    if (my_reply->barrier_id == xlastpos)
      if (req.complex_request.accepts_reply( req, *my_reply)) {
	req.reply( *my_reply);
	my_reply= NULL;
	return +1;
      }
     // this reply is not the one we are waiting for, discard it
    
    delete my_reply;
    my_reply= NULL;    
  }

  omni_thread::get_time( 	&sec, &nsec, 
				timeout.seconds, timeout.microseconds*1000);
  if (req.cancelled()) return -1;

  while (1) {
    int wait_ret;

#ifdef OMNIORB
    wait_ret= cond_reply_arrived.timedwait( sec, nsec);
#else
    wait_ret= cond_reply_arrived.timed_wait( sec, nsec);
#endif
    if (my_reply) {
      if (req.complex_request.accepts_reply( req, *my_reply)) {
	req.reply( *my_reply);
	my_reply= NULL;
	return +1;
      }
      // this reply is not the one we are waiting for, discard it
#ifdef DEBUG    
      cout << "Unexpected reply" << endl;
#endif
      delete my_reply;
      my_reply= NULL;    
    }
    if (req.cancelled()) return -1;
#ifdef OMNIORB
    if (wait_ret == 0 || cancel_reply) return 0;
#else
    if (wait_ret == ETIMEDOUT || cancel_reply) return 0;
#endif
  }
}
 
void Sched_Barrier::reply_arrived( Sched_Reply *reply) {
  auto_lock local_lock( *this);
  if (my_reply) delete my_reply;
  my_reply= reply;

  cond_reply_arrived.signal();
}

