//
//  File:         sensor_rep_i.cc
//  Author:       Andrzej Uszok (uszok@ics.agh.edu.pl)
//  Date Created: 03.10.97
//
// General Description:
//   Implementation file for the Sensor_Rep_I class
//

#define EXCEPTIONS

#include "sensor_rep_i.h"
#include "Naming.hh"
#include "db_utils.hh"

//
// 		Constructor
//              ___________

   Sensor_Rep_I::Sensor_Rep_I ( 
#ifdef ObjectStore
    	os_database* db_p
#elif defined FilePersistency
    	const char *repository_name_p
#endif
    ) : Basic_Rep_I<Sensor_Types, S_SDO_Types>(hash_sensor,
#ifdef ObjectStore
    		db_p
#elif defined FilePersistency
    		repository_name_p
#endif
        ) 
{
  	_instance = this;
  	create_corba_objects();
}


void 
   Sensor_Rep_I::init (
	int cold_start_p
   ) 
{
  Basic_Rep_I<Sensor_Types, S_SDO_Types>::init(cold_start_p);

  try {
    Bootstrap::register_server_in_NS( Bootstrap::SRV_Sensor_Repository(), my_rep);

    register_callback(Bootstrap::SRV_Scheduler());
    ps_dev= new ProcessControlImpl( "sensor_repository", this);
  }
  catch (CORBA::Exception &ex) {
    cerr << "Caught exception while contacting server manager " << ex << endl;
    exit(1);
  }
}

void 
   Sensor_Rep_I::init_corba_object ( 
	Sensor_I *item_p
   ) 
{
	 // create the TIE (CORBA) object for the C++ object
#ifdef ORBIX
#ifdef ObjectStore
	ObjectStoreTIE_ABng_Core_System_Sensor(Sensor_I)* new_sensor_tie =
		new(my_db, ObjectStoreTIE_ABng_Core_System_Sensor(Sensor_I)::get_os_typespec())
		ObjectStoreTIE_ABng_Core_System_Sensor(Sensor_I) (item_p);
#else
	TIE_ABng_Core_System_Sensor(Sensor_I)* new_sensor_tie = 
		new TIE_ABng_Core_System_Sensor(Sensor_I) (item_p);
#endif
#elif defined OMNIORB
  ABng_Core_System::Sensor_ptr new_sensor_tie=
    new TIE_CLASS(ABng_Core_System,Sensor)<Sensor_I,0>(item_p);
  CONNECT_IMPL( ABng_Core_System::Sensor_var, new_sensor_tie);
	
#endif
  item_p->set_corba_object(new_sensor_tie);
}

void 
   Sensor_Rep_I::create_corba_objects (
   ) 
{
#ifdef ORBIX
#ifdef FilePersistency
	// create a TIE for the object
	my_rep = new TIE_ABng_Core_Management_Sensor_Rep(Sensor_Rep_I) (this);
#endif /* FilePersistency */
#elif defined OMNIORB
  my_rep =
    new TIE_CLASS(ABng_Core_Management,Sensor_Rep)<Sensor_Rep_I,0>(this);
  CONNECT_IMPL( ABng_Core_Management::Sensor_Rep_var,my_rep);
#endif

}

Sensor_Rep_I* Sensor_Rep_I::_instance = 0;

Sensor_Rep_I* 
   Sensor_Rep_I::instance(
   )
{
   	return _instance;
}

ABng_Core_System::Scheduler_var 
   Sensor_Rep_I::scheduler (
   ) 
{ 
	auto_lock local_lock( *this);
	return r_scheduler; 
}

//
// 		Destructor
//		__________

   Sensor_Rep_I::~Sensor_Rep_I()
{
  	destroy_corba_objects();
	delete ps_dev;
}

void 
   Sensor_Rep_I::destroy_corba_objects (
   ) 
{ 
	CORBA::release(my_rep);
}

//
// Private functions
//

void 
   Sensor_Rep_I::reference_resolved ( 
	const ABng_Server_Manager::Name &name,
	CORBA::Object_var server_ref,
	ABng_Server_Manager::uid callback_id
   ) 
{
	auto_lock local_lock(*this);

  	if (name == Bootstrap::SRV_Scheduler())
		try {
      			r_scheduler = ABng_Core_System::Scheduler::_narrow(server_ref);
    		}
    		catch (...) {
      			r_scheduler = ABng_Core_System::Scheduler::_nil();
    		}
}

DB_Header* 
   Sensor_Rep_I::create_db_header (
   ) 
{
  	return new DB_Header("Sensor Repository", "1.0", "\
#       <data_record>::=  	<item_id> <sensor_id>\n\
#\n\
#       <sensor_id>::=		<sensor_number> <networ_number>\n\
#				<group_name>\n\
#\n\
#       <sensor_number>::=	<decimal_numbers>\n\
#       <network_number>::=	<decimal_numbers>\n\
#       <group_name>::=		<string>\n\
#\n\
");
}

unsigned 
   Sensor_Rep_I::hash_sensor(
	const Sensor_I& sensor_i_p
   )
{
	ABng_Core_System::Sensor::Description sensor_descr = 
		sensor_i_p.get_descr();

	return 256 * sensor_descr.id.network_num + sensor_descr.id.sensor_num;
}

void
   Sensor_Rep_I::gather_sensor(
	Sensor_I *current_sensor,
	void* apply_data
   )
{
	// project apply_data (void*) to Sensors*
	ABng_Core_System::Sensors *collected_sensors =
		(ABng_Core_System::Sensors*)apply_data;

	// increase the length of the container
	collected_sensors->length(collected_sensors->length() + 1);

	// put duplicated sensor (CORBA object) into the container
	(*collected_sensors)[collected_sensors->length() - 1] =
		ABng_Core_System::Sensor::_duplicate(current_sensor->get_corba_object());
}

void
   Sensor_Rep_I::match_sensor(
	Sensor_I *current_sensor,
	void* apply_data
   )
{
	const ABng_Core_System::Sensor::Pattern& data_record_pattern = 
		((struct MatchSensorApplyDataStruct*)apply_data)->_data_record_pattern;

	ABng_Core_System::Sensor::Description current_sensor_descr;

	// check general patten matching rule
	if (data_record_pattern.is_any){
		// any sensor matches this pattern
		gather_sensor(current_sensor, 
			((struct MatchSensorApplyDataStruct*)apply_data)->_found_sensors);
		return;
	}

	current_sensor_descr = current_sensor->get_descr();

	// check matching of item_id
	if (!data_record_pattern.item_id_pattern.is_any &&
	    (data_record_pattern.item_id_pattern.value != current_sensor_descr.item_id))
		// this sensor item_id does not match pattern 
		return;

	// check matching of id
	if (!data_record_pattern.id_pattern.is_any){
		// check matching of id's sensor_number
		if (!data_record_pattern.id_pattern.sensor_num_pattern.is_any 
		    && data_record_pattern.id_pattern.sensor_num_pattern.value 
		    != current_sensor_descr.id.sensor_num)
			// this sensor id.sensor_number does not match pattern
			return;

		// check matching of id's network_number
		if (!data_record_pattern.id_pattern.network_num_pattern.is_any 
		    && data_record_pattern.id_pattern.network_num_pattern.value
		    != current_sensor_descr.id.network_num)
			// this sensor id.network_number does not match pattern
			return;

		// check matching of id's network_group_name (string !!!)
		if (!data_record_pattern.id_pattern.network_grp_name_pattern.is_any)
			if (!ABng_String::match_string(
			    data_record_pattern.id_pattern.network_grp_name_pattern.value, 
			    current_sensor_descr.id.group_name))
				// this sensor id.group_name does not match pattern
				return;
	}

	// this sensor matches the pattern
	gather_sensor(current_sensor, 
		((struct MatchSensorApplyDataStruct*)apply_data)->_found_sensors);
}

void 
   Sensor_Rep_I::construct_data_object (
	ABng_Core_System::System_Static_Data_Observer::System_Static_Data_Object& data_object_p,
	ABng_Core_System::Sensor_ptr item_p
   ) 
{
	// construct union to store reference of added entity
	data_object_p.sensor_obj(ABng_Core_System::Sensor::_duplicate(item_p));
}

void 
   Sensor_Rep_I::collect_values ( 
	Sensor_I *lt, 
	ABng_Core_System::Sensor::Value_Description& ld
   ) 
{
	ld = lt->get_descr();
}

//
// Implementation of IDL operations of the System_Static_Data_Observed_R interface
//                                         _____________________________

ABng_Core_System::System_Static_Data_Observed_R::Observer_Entries*
   Sensor_Rep_I::system_static_data_observer_list (
	ENV_DEF1(env)
   ) 
		throw (CORBA::SystemException)
{	
	return observer_list(USE_ENV1(env));
}

ABng_Core_System::System_Static_Data_Observed_E::System_Static_Data_Observers*
   Sensor_Rep_I::add_system_static_data_observer_list (
	ENV_DEF1(env)
   ) 
		throw (CORBA::SystemException)
{
	return add_observer_list(USE_ENV1(env));
}
                        
void 
   Sensor_Rep_I::attach_system_static_data_observer (
	ABng_Core_System::System_Static_Data_Observer_ptr obs,  
	const Domain_Repository::Repository_Item_Ids& observed_id_list
	ENV_DEF(env)
   )
		throw (CORBA::SystemException, 
                       Domain_Repository::Unknown_Repository_Item_Ids)
{
	attach_observer(obs, observed_id_list USE_ENV(env));
}

void 
   Sensor_Rep_I::detach_system_static_data_observer (
	ABng_Core_System::System_Static_Data_Observer_ptr obs, 
	const Domain_Repository::Repository_Item_Ids& observed_id_list
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException, 
                       Domain_Repository::Unknown_Repository_Item_Ids, 
                       Domain_Repository::Unknown_Observer,
		       Domain_Repository::Unmatched_Observer_Item_Ids)
{
	detach_observer(obs, observed_id_list USE_ENV(env));
}

void 
   Sensor_Rep_I::attach_add_system_static_data_observer (
	ABng_Core_System::System_Static_Data_Observer_ptr obs  
	ENV_DEF(env)
   )
		throw (CORBA::SystemException)
{
	attach_add_observer(obs USE_ENV(env));
}

void 
   Sensor_Rep_I::detach_add_system_static_data_observer (
	ABng_Core_System::System_Static_Data_Observer_ptr obs 
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException, 
                       Domain_Repository::Unknown_Observer)
{
	detach_add_observer(obs USE_ENV(env));
}


// Suplementary operation

ABng_Core_System::System_Static_Data_Observed_E::System_Static_Data_Observers*
   Sensor_Rep_I::system_static_data_observer_list (
	Domain_Repository::Repository_Item_Id item_id_p
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException)
{
	return get_item_observers(item_id_p USE_ENV(env));
}

//
// Implementation of IDL operations of the Sensor_Rep interface
//                                         __________

ABng_Core_System::Sensors* 
   Sensor_Rep_I::sensor_list (
	ENV_DEF1(env)
   ) 
		throw (CORBA::SystemException)
{
	return item_list(USE_ENV1(env));
}
                
ABng_Core_System::Sensor_ptr 
   Sensor_Rep_I::add (
	const ABng_Core_System::Sensor::Description& data_record
	ENV_DEF(env)
   )
		throw (CORBA::SystemException, 
                       Domain_Repository::Duplicate_Data_Record)
{
	return Basic_Rep_I<Sensor_Types, S_SDO_Types>::add(data_record USE_ENV(env));
}

void 
   Sensor_Rep_I::remove (
	ABng_Core_System::Sensor_ptr object_to_remove 
	ENV_DEF(env)
   )
		throw (CORBA::SystemException, 
                       Domain_Repository::Unknown_Object_Ref)
{
	Basic_Rep_I<Sensor_Types, S_SDO_Types>::remove(object_to_remove USE_ENV(env));
}

ABng_Core_System::Sensors* 
   Sensor_Rep_I::find (
	const ABng_Core_System::Sensor::Pattern& data_record_pattern
	ENV_DEF(env)
   )
		throw (CORBA::SystemException,
                       Domain_Repository::Search_Not_Implemented)
{
  	auto_lock local_lock(*this);

  	return private_find(data_record_pattern USE_ENV(env));
}

ABng_Core_System::Sensors* 
   Sensor_Rep_I::private_find (
	const ABng_Core_System::Sensor::Pattern& data_record_pattern
	ENV_DEF(env)
   )
		throw (CORBA::SystemException,
                       Domain_Repository::Search_Not_Implemented)
{
	// create a structure consisting of a pointer to the container
	//  and a reference of the pattern
	MatchSensorApplyDataStruct apply_data = MatchSensorApplyDataStruct(
		(ABng_Core_System::Sensors*)new
 	ABng_Core_System::Sensors(my_item_list.entries()),
		data_record_pattern);

	// browse through hash table to collect sensors
	my_item_list.apply(match_sensor, (void*)&apply_data);

	// pointer to found_sensors was stored inside apply_data thus return it
	return apply_data._found_sensors;
}

ABng_Core_System::Sensor::Value_Descriptions* 
   Sensor_Rep_I::find_values (
	const ABng_Core_System::Sensor::Pattern& data_record_pattern
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException,
                       Domain_Repository::Search_Not_Implemented)
{
	return Basic_Rep_I<Sensor_Types, S_SDO_Types>::find_values(data_record_pattern USE_ENV(env));
}

void 
   Sensor_Rep_I::update (
	ABng_Core_System::Sensor_ptr object_to_update, 
	const ABng_Core_System::Sensor::Description& data_record_pattern
	ENV_DEF(env)
   ) 
		throw (CORBA::SystemException, 
		       Domain_Repository::Unknown_Object_Ref,
		       Domain_Repository::Duplicate_Data_Record)
{
	Basic_Rep_I<Sensor_Types, S_SDO_Types>::update(object_to_update, data_record_pattern USE_ENV(env));
}
