//
// $Id: server_mgr.hh,v 1.4 1998/11/30 15:04:21 mencnar Exp $
//
// $Log: server_mgr.hh,v $
// Revision 1.4  1998/11/30 15:04:21  mencnar
// Added support for checking for status of garbage collector
//
// Revision 1.3  1998/07/29 14:04:28  mencnar
// "Server_Status" implementation
//
// Revision 1.2  1998/06/18 10:35:54  mencnar
// moved to CVS
//
// Revision 1.1  1998/03/05 13:53:58  mencnar
// Initial revision
//
//
//-----------------------------------------------------------------------------
// 
// Author:      Daniel Mencnarowski  (mencnar@ics.agh.edu.pl)
//
// Description: Header file for
//              implementation of ABng_Server_Manager::Register interface.
//              Keeps track of all running servers in ABng system.
//

#ifndef __server_mgr_hh__
#define __server_mgr_hh__

#include "abng_server_manager.hh"
#include "compatibility.hh"
#include "bootstrap_base.hh"
#include "abng_guarded.hh"
#include "domain_rep.hh"
#include "db_utils.hh"
#include "process_status.hh"
#include <rw/tvhasht.h>
#include <rw/tphasht.h>
#include <rw/tvdlist.h>
#include <rw/rwstring.h>

unsigned hash_server_name( const ABng_Server_Manager::Name &);
int operator == ( const ABng_Server_Manager::Name &,
		  const ABng_Server_Manager::Name &);

//--------------------------------------------------------------------Server_Entry--
class Server_Entry {
public:
  ABng_Server_Manager::uid item_id;

  CORBA::Object_var server_reference;
#ifdef ORBIX
#if CORBA_ORBIX_VER < 23
  RWCString stringified_server;
#endif
#endif
  ABng_Server_Manager::Name server_name;

  Server_Entry() {}

  Server_Entry( ABng_Server_Manager::uid _item_id,
	        CORBA::Object_var &_server_reference,
		const ABng_Server_Manager::Name &_name);

  static unsigned hash( const Server_Entry &);
  int operator == ( const Server_Entry &) const;
  int operator >= ( const Server_Entry &) const;
  int is_public();

  static int name_is_public( const ABng_Server_Manager::Name &name);
};

//------------------------------------------------------------------Callback_Entry--
class Callback_Entry {
public:
  ABng_Server_Manager::uid item_id;

  ABng_Server_Manager::Callback_var callback_object;
#ifdef ORBIX
#if CORBA_ORBIX_VER < 23
  RWCString stringified_callback;
#endif
#endif
  ABng_Server_Manager::Name server_name;
  int context_entry;
  int public_entry;

  Callback_Entry() {}
             
  Callback_Entry( ABng_Server_Manager::uid _item_id,
		  ABng_Server_Manager::Callback_var &_callback_object,
		  const ABng_Server_Manager::Name &_server_name,
		  int _context_entry,
		  int _public_entry);


  static unsigned hash( const Callback_Entry &);
  int operator == ( const Callback_Entry &) const;
};

//------------------------------------------------------------------Notifier_Entry--
class Notifier_Entry : public ABng_S_Guarded {
  int refcount;
public:
  ABng_Server_Manager::Name server_name;
  
  
  Notifier_Entry( const ABng_Server_Manager::Name &_server_name) :
    server_name( _server_name),
    refcount( 1) {}

  void duplicate() { refcount++; }
  int release() { return --refcount; }

  static unsigned hash( const Notifier_Entry &);
  int operator == ( const Notifier_Entry &) const;
};

class Server_Manager;


//----------------------------------------------------------------------Callback_DB--
class Callback_DB : public ABng_Guarded,
                    public ServerStatusInterface {
public:
  typedef int ( Callback_DB::* callback_fun)( Callback_Entry &info,
					      Server_Entry &server);


private:
  TransactionLOG &tlog;

  RWTValHashTable< Callback_Entry> db;
  RWTValHashTable< Server_Entry> server_db;
  RWTPtrHashTable< Notifier_Entry> active_notifiers;

  unsigned long unique_id;
  RWCString file_name;
  Bootstrap_base &base;
  int gc_sleep_time;

  RWTValDlist< Callback_Entry>
    Callback_DB::gather_callbacks( ABng_Server_Manager::Name &name);

  RWTValDlist< Callback_Entry>
    gather_all_callbacks();

  RWTValDlist< Callback_Entry>
    Callback_DB::private_gather_callbacks( ABng_Server_Manager::Name &name);

  RWTValDlist< Server_Entry>
    gather_all_servers();

  RWTValDlist< Server_Entry> 
    Callback_DB::gather_servers( ABng_Server_Manager::Name &name,
				 int check_NS,
				 int public_callback) ;
  RWTValDlist< Server_Entry> 
    Callback_DB::private_gather_servers( ABng_Server_Manager::Name &name,
					 int check_NS,
					 int public_callback);

  RWTValDlist< Server_Entry>
    Callback_DB::gather_servers_in_context( ABng_Server_Manager::Name &name,
					    int check_NS,
					    int public_callback);

  RWTValDlist< Server_Entry>
    Callback_DB::private_gather_servers_in_context( ABng_Server_Manager::Name &name,
						    int check_NS,
						    int public_callback);

  static void static_garbage_collector( void *_params);
  void sweep_db();
  void garbage_collector();

  int private_remove( Callback_Entry &callback_entry,
				   int unique);
  int private_remove_server( Server_Entry &entry, 
					  int unique);
  void launch_server_broadcast( 
			       Server_Entry &server_entry,
			       RWTValDlist< Callback_Entry> &callbacks,
			       callback_fun callback);
  static void static_launch_server_broadcast( void *_params);
  void server_broadcast( Server_Entry &server_entry,
			 RWTValDlist< Callback_Entry> &callbacks,
			 callback_fun callback);
  void launch_callback_broadcast( 
				 Callback_Entry &callback_entry,
				 RWTValDlist< Server_Entry> &servers,
				 callback_fun callback);
  static void static_launch_callback_broadcast( void *_params);
  void callback_broadcast( Callback_Entry &callback_entry,
			   RWTValDlist< Server_Entry> &servers,
			   callback_fun callback);
  int check_server_entry( Callback_Entry &,
			  Server_Entry &server_entry);
  int check_callback_entry( Callback_Entry &callback_entry,
			    Server_Entry &);
  int send_callback_startup( Callback_Entry &callback_entry,
			     Server_Entry &server_entry);
  int send_callback_shutdown( Callback_Entry &callback_entry,
			      Server_Entry &server_entry);
  void save();

public:

  Callback_DB( const char *_file_name,
	       Bootstrap_base &_base,
	       TransactionLOG &_tlog,
	       int _gc_sleep_time) : 
    unique_id( 1),
    db( Callback_Entry::hash),
    server_db( Server_Entry::hash),
    active_notifiers( Notifier_Entry::hash),
    file_name( _file_name),
    base( _base),
    gc_sleep_time( _gc_sleep_time),
    tlog( _tlog) {
  }

  void start_GC();
  void restore();

  void insert( Callback_Entry &callback_entry);
  int remove( Callback_Entry &callback_entry,
			   int unique);
  int insert_server( Server_Entry &entry);
  int remove_server( Server_Entry &entry, 
				  int unique);

  virtual void get_server_status( ostream &);

  friend class Notifier_Lock;
};

//-----------------------------------------------------------------Notifier_Lock--
//
class Notifier_Lock {
  Callback_DB &db;
  Notifier_Entry *found;
  ABng_Server_Manager::Name server_name;
public:
  enum callback_info { STARTUP_INFO, SHUTDOWN_INFO, 
		       CHECK_INFO_SERVER, CHECK_INFO_CALLBACK };

  Notifier_Lock( Callback_DB &_db,
		 Server_Entry &server_entry,
		 callback_info ci,
		 CORBA::Object_ptr call_obj);
  ~Notifier_Lock();
};

//-----------------------------------------------------------------Server_Params--
//
class Server_Params {
public:
  Callback_DB &callback_db;
  Server_Entry server_entry;
  RWTValDlist< Callback_Entry> callbacks;
  Callback_DB::callback_fun callback;

  Server_Params( Callback_DB &_callback_db,
		 Server_Entry &_server_entry,
		 RWTValDlist< Callback_Entry> &_callbacks,
		 Callback_DB::callback_fun _callback) :
    callback_db( _callback_db),
    server_entry( _server_entry),
    callbacks( _callbacks),
    callback( _callback) {}
};

//---------------------------------------------------------------Callback_Params--
//
class Callback_Params {
public:
  Callback_DB &callback_db;
  Callback_Entry callback_entry;
  RWTValDlist< Server_Entry> servers;
  Callback_DB::callback_fun callback;

  Callback_Params( Callback_DB &_callback_db,
		   Callback_Entry &_callback_entry,
		   RWTValDlist< Server_Entry> &_servers,
		   Callback_DB::callback_fun _callback) :
    callback_db( _callback_db),
    callback_entry( _callback_entry),
    servers( _servers),
    callback( _callback) {}
};

//----------------------------------------------------------------Server_Manager--


class Server_Manager : 

  public Bootstrap_base {

private:
  Callback_DB db;
  TransactionLOG tlog;

  ABng_Server_Manager::Register_ptr register_object;
  ABng_Server_Manager::Register_Manager_ptr register_manager_object;

public:
  Server_Manager( CORBA::ORB_ptr,
		  int gc_sleep_time);


  ServerStatusInterface *server_status_interface() {
    return &db;
  }

  void register_server( CORBA::Object_ptr server_reference,
			const ABng_Server_Manager::Name &name
			ENV_DECL( env) );

  void unregister_server( const ABng_Server_Manager::Name &name
			  ENV_DECL( env) );

  ABng_Server_Manager::uid 
    register_callback( ABng_Server_Manager::Callback_ptr callback_object,
		       const ABng_Server_Manager::Name& server_name
		       ENV_DECL( env) );

  ABng_Server_Manager::uid 
    register_restricted_callback( 
		       ABng_Server_Manager::Callback_ptr callback_object,
		       const ABng_Server_Manager::Name& server_name
		       ENV_DECL( env) );

  ABng_Server_Manager::uid 
    register_context_callback( ABng_Server_Manager::Callback_ptr callback_object,
			       const ABng_Server_Manager::Name& context_name
			       ENV_DECL( env) );

  ABng_Server_Manager::uid 
    register_restricted_context_callback( 
			    ABng_Server_Manager::Callback_ptr callback_object,
			    const ABng_Server_Manager::Name& context_name
			    ENV_DECL( env) );

  void unregister_callback( ABng_Server_Manager::uid identifier
			    ENV_DECL( env) );

  void unregister_restricted_callback( ABng_Server_Manager::uid identifier
				       ENV_DECL( env) );

  ABng_Server_Manager::Time get_current_time( ENV_DECL1( env));
  ABng_Server_Manager::ip_address* get_ip_address( ENV_DECL1( env));

  ABng_Server_Manager::Register_ptr reg( ENV_DECL1(env)) { 
    return ABng_Server_Manager::Register::_duplicate
      ((ABng_Server_Manager::Register_ptr) register_object);
  }
};

#ifdef ORBIX
#ifdef ObjectStore
DEF_ObjectStoreTIE_ABng_Server_Manager_Register(Server_Manager)
DEF_ObjectStoreTIE_ABng_Server_Manager_Register_Manager(Server_Manager)
#else
DEF_TIE_ABng_Server_Manager_Register(Server_Manager)
DEF_TIE_ABng_Server_Manager_Register_Manager(Server_Manager)
#endif
#endif


#endif // __server_mgr_hh__

//










